"""
Relevance, Full-stack search engine, core module.

Relevance is a software suite that aims to address the common and recurring problem
of ingesting content, delivering it, searching it with relevant results, analyzing
its usage by the end-users, acting on the analytic data to provide the best results
possible.

Its primary goals are rapid deployment, extensibility and ease of use.
"""

# Package version
__version__ = '0.4.2'

# Author information
__author__ = 'Francis Lacroix'

# License information
__license__ = 'MIT'
__copyright__ = 'OverrideLogic 2016'
